const shader = `// A fragment shader which lights textured geometry with point lights.

// Lights from a storage buffer binding.
struct PointLight {
  position : vec3f,
  color : vec3f,
}

struct LightStorage {
  pointCount : u32,
  point : array<PointLight>,
}
@group(0) @binding(0) var<storage> lights : LightStorage;

// Texture and sampler.
@group(1) @binding(0) var baseColorSampler : sampler;
@group(1) @binding(1) var baseColorTexture : texture_2d<f32>;

// Function arguments are values from the vertex shader.
@fragment
fn fragmentMain(@location(0) worldPos : vec3f,
                @location(1) normal : vec3f,
                @location(2) uv : vec2f) -> @location(0) vec4f {
  // Sample the base color of the surface from a texture.
  let baseColor = textureSample(baseColorTexture, baseColorSampler, uv);

  let N = normalize(normal);
  var surfaceColor = vec3f(0);

  // Loop over the scene point lights.
  for (var i = 0u; i < lights.pointCount; i++) {
    let worldToLight = lights.point[i].position - worldPos;
    let dist = length(worldToLight);
    let dir = normalize(worldToLight);

    // Determine the contribution of this light to the surface color.
    let radiance = lights.point[i].color * (1 / pow(dist, 2));
    let nDotL = max(dot(N, dir), 0);

    // Accumulate light contribution to the surface color.
    surfaceColor += baseColor.rgb * radiance * nDotL;
  }

  // Return the accumulated surface color.
  return vec4(surfaceColor, baseColor.a);
}`

/**
 * 
      3.1 Parsing
        Remove comments:

            Replace the first comment with a space code point (U+0020).

            Repeat until no comments remain.

        Find template lists, using the algorithm in § 3.9 Template Lists.

        Parse the whole text, attempting to match the translation_unit grammar rule. Parsing uses a LALR(1) parser (one token of lookahead) [DeRemer1969], with the following customization:

            Tokenization is interleaved with parsing, and is context-aware. When the parser requests the next token:

                Consume and ignore an initial sequence of blankspace code points.

                If the next code point is the start of a template list, consume it and return _template_args_start.

                If the next code point is the end of a template list, consume it and return _template_args_end.

                Otherwise:

                    A token candidate is any WGSL token formed from the non-empty prefix of the remaining unconsumed code points.

                    The token returned is the longest token candidate that is also a valid lookahead token for the current parser state. [VanWyk2007]


 */

