import { any, char, oneOf, digit, many, map, maybe, not, seq, tag } from 'kuebiko'
import { Token } from './token.js'

export class Literal extends Token {
  constructor(value) {
    super(value)
  }
}

export class BooleanLiteral extends Token {
  static from(value) {
    return new BooleanLiteral(value)
  }
}

export class IntegerLiteral extends Token {
  _dataType
  _system

  constructor(value, type, system) {
    super(value)
    this._dataType = type
    this._system = system
  }
}

class Signed {
  static Signed = new Signed(true)
  static Unsigned = new Signed(false)
  _value

  constructor(value) {
    this._value = value
  }
}

export const iu = any(
  map(() => Signed.Signed, char('i')),
  map(() => Signed.Unsigned, char('u'))
)

const decimal = char('.')
const sign = oneOf('+-')
const e = oneOf('eE')
const p = oneOf('pP')
const fh = oneOf('fh')

export const zero = char('0')
export const digits = many(digit)

export const booleanLiteral = any(
  map(BooleanLiteral.from, tag('true')),
  map(BooleanLiteral.from, tag('false')),
)

export const decimalIntLiteral = any(
  seq(
    digits,
    maybe(iu)
  )
)

const hexDigit = any(digit, oneOf('abcdef'), oneOf('ABCDEF'))

const hexPrefix = seq(zero, oneOf('xX'))

const hexIntLiteral = seq(
  hexPrefix,
  hexDigit,
  maybe(iu)
)

export const integerLiteral = any(
  zero,
  decimalIntLiteral,
  hexIntLiteral
)


const nonzeroDigit = oneOf('123456789')

const many1 = parser => seq(parser, many(parser))

const exponent = seq(
  e,
  sign,
  many1(digit)
)

const fhDecimalFloatLiteral = seq(
  any(
    zero,
    seq(
      nonzeroDigit,
      digits
    )
  ),
  fh
)

const eDecimalFloatLiteral = seq(
  digits,
  exponent,
  fh
)

const cDecimalFloatLiteral = seq(
  any(
    seq(
      digits,
      decimal,
      many1(digit)
    ),

    seq(
      many1(digit),
      decimal,
      digits
    )
  ),

  maybe(seq(exponent, fh)),
)

const decimalFloatLiteral = any(
  fhDecimalFloatLiteral,
  eDecimalFloatLiteral,
  cDecimalFloatLiteral
)

const sHexFloatLiteral = seq(
  hexPrefix,
  any(
    seq(
      many(hexDigit),
      decimal,
      many1(hexDigit)
    ),

    seq(
      many1(hexDigit),
      decimal,
      many(hexDigit)
    )
  ),

  maybe(
    seq(
      p,
      maybe(sign),
      many1(digit),
      fh
    )
  )
)

const zHexFloatLiteral = seq(
  hexPrefix,
  many1(hexDigit),
  p,
  maybe(sign),
  many1(digit),
  maybe(fh)
)
const hexFloatLiteral = any(
  sHexFloatLiteral,
  zHexFloatLiteral
)

export const floatingPointLiteral = any(
  decimalFloatLiteral,
  hexFloatLiteral
)


export const numericLiteral = any(
  integerLiteral,
  floatingPointLiteral
)

export const literal = any(
  numericLiteral,
  booleanLiteral
)
